#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/mount.h>
#include <sys/stat.h>
#include <sys/poll.h>
#include <errno.h>
#include <dirent.h>
#include <sys/stat.h>

#define LOG_TAG "RIL"
#include <utils/Log.h>
#include <cutils/properties.h>

#if 1 //quectel PLATFORM_VERSION >= "4.2.2"
#ifndef LOGD
#define LOGD ALOGD
#endif
#ifndef LOGE
#define LOGE ALOGE
#endif
#ifndef LOGI
#define LOGI ALOGI
#endif
#endif

static int ql_ip_up(void) {
    char *dns1 = getenv("DNS1");
    char *dns2 = getenv("DNS2");
    char *iplcocal = getenv("IPLOCAL");
    char *ipremote = getenv("IPREMOTE");

    //property_set("net.interfaces.defaultroute", "ppp0");
    property_set("net.ppp0.dns1", dns1 ? dns1 : "");
    property_set("net.ppp0.dns2", dns2 ? dns2 : "");
    property_set("net.ppp0.local-ip", iplcocal ? iplcocal : "");
    property_set("net.ppp0.remote-ip", ipremote ? ipremote : "");
    return 0;
}

static int ql_ip_down(void) {
    //property_set("net.interfaces.defaultroute", "");
    property_set("net.ppp0.dns1", "");
    property_set("net.ppp0.dns2", "");
    property_set("net.ppp0.local-ip", "");
    property_set("net.ppp0.remote-ip", "");
    return 0;
}        

static char *pppd_args[] = {"pppd", "$DEVNAME", "115200", "nodetach", "nolock", "debug", "dump", "user", "$USER", "password", "$PASSWORD",
    "nocrtscts", "modem", "hide-password", "usepeerdns", "noipdefault", "novj", "novjccomp", "noccp", "defaultroute", "ipcp-accept-local", "ipcp-accept-remote",
    "ipcp-max-failure", "10",
    "connect", "/etc/ppp/init.quectel-pppd chat connect",
    "disconnect","/etc/ppp/init.quectel-pppd chat disconnect",
    NULL
};

#define MAX_PATH 256
#define USBID_LEN 4
struct ql_usb_id_struct {
    unsigned short vid;
    unsigned short pid;
    unsigned short at_inf;
    unsigned short ppp_inf;
    
};
static const struct ql_usb_id_struct ql_usb_id_table[] = {
    {0x05c6, 0x9003, 2, 3}, //UC20
    {0x05c6, 0x9090, 2, 3}, //UC15
    {0x05c6, 0x9215, 2, 3}, //EC20
    {0x1519, 0x0331, 6, 0}, //UG95
    {0x1519, 0x0020, 6, 0}, //UG95
};
#define USB_AT_INF 0
#define USB_PPP_INF 1
#ifndef ARRAY_SIZE
#define ARRAY_SIZE(a) (sizeof(a) / sizeof(a[0]))
#endif
static int is_usb_match(unsigned short vid, unsigned short pid) {
    size_t i;
    for (i = 0; i < ARRAY_SIZE(ql_usb_id_table); i++) {
        if (vid == ql_usb_id_table[i].vid) {
            if (pid == 0x0000) //donot check pid
                return 1;
            else if (pid == ql_usb_id_table[i].pid)
                return 1;
        }
    }
    return 0;
}

static int idusb2hex(char idusbinfo[USBID_LEN]) {
    int i;
    int value = 0;
    for (i = 0; i < USBID_LEN; i++) {
        if (idusbinfo[i] < 'a')
            value |= ((idusbinfo[i] - '0') << ((3 - i)*4));
        else
             value |= ((idusbinfo[i] - 'a' + 10) << ((3 - i)*4));
    }
    return value;
}

static int ql_get_ttyname(int usb_interface, char *out_ttyname) {
    struct dirent* ent = NULL;  
    DIR *pDir;  
    char dir[MAX_PATH], filename[MAX_PATH];
    struct stat statbuf;
    int idVendor = 0, idProduct = 0;
    int fd;
    int find_usb_device = 0;
    size_t i;

    out_ttyname[0] = '\0';
    dir[0] = '\0';
    strcat(dir, "/sys/bus/usb/devices");
    if ((pDir = opendir(dir)) == NULL)  {  
        LOGE("Cannot open directory:%s/", dir);  
        return -1;  
    }  

    while ((ent = readdir(pDir)) != NULL)  {
        sprintf(filename, "%s/%s", dir, ent->d_name);  
        lstat(filename, &statbuf);
        if (S_ISLNK(statbuf.st_mode))  {
            char idusbinfo[USBID_LEN+1] = {0};
            
            idVendor = idProduct = 0x0000;
            sprintf(filename, "%s/%s/idVendor", dir, ent->d_name);
            fd = open(filename, O_RDONLY);
            if (fd > 0) {
                if (4 == read(fd, idusbinfo, USBID_LEN))
                    idVendor = idusb2hex(idusbinfo);
                close(fd);
            }
            if (!is_usb_match(idVendor, idProduct))
                continue;

            sprintf(filename, "%s/%s/idProduct", dir, ent->d_name);
            fd = open(filename, O_RDONLY);
            if (fd > 0) {
                if (4 == read(fd, idusbinfo, USBID_LEN))
                    idProduct = idusb2hex(idusbinfo);
                close(fd);
            }
            if (!is_usb_match(idVendor, idProduct))
                continue;
        
            LOGD("find vid=0x%04x, pid=0x%04x", idVendor, idProduct);
            find_usb_device = 1;
            break;
        }
    }
    closedir(pDir);

    if (usb_interface > 0xf)
        return idProduct;

    for (i = 0; i < ARRAY_SIZE(ql_usb_id_table); i++) {
        if ((idVendor == ql_usb_id_table[i].vid) && (idProduct == ql_usb_id_table[i].pid)) {
            if (usb_interface == USB_AT_INF)
                usb_interface = ql_usb_id_table[i].at_inf;
            else if (usb_interface == USB_PPP_INF)
                usb_interface = ql_usb_id_table[i].ppp_inf;
        }
    }

    if (find_usb_device) {
        char usb_inf_path[20];
        sprintf(usb_inf_path, "/%s:1.%d", ent->d_name, usb_interface);
        strcat(dir, usb_inf_path);
        if ((pDir = opendir(dir)) == NULL)  {  
            LOGE("Cannot open directory:%s/", dir);  
            return -1;
        }
                       
        while ((ent = readdir(pDir)) != NULL)  {
            if (strncmp(ent->d_name, "tty", 3) == 0) {
                LOGD("find vid=0x%04x, pid=0x%04x, tty=%s", idVendor, idProduct, ent->d_name);
                strcpy(out_ttyname, ent->d_name);
            break;
            } 
        }
        closedir(pDir); 
    }

    if (strcmp(out_ttyname, "tty") == 0) { //find tty not ttyUSBx or ttyACMx
        strcat(dir, "/tty");
        if ((pDir = opendir(dir)) == NULL)  {  
            LOGE("Cannot open directory:%s/", dir);  
            return -1;
        }
        
        while ((ent = readdir(pDir)) != NULL)  {
            if (strncmp(ent->d_name, "tty", 3) == 0) {
                LOGD("find vid=0x%04x, pid=0x%04x, tty=%s", idVendor, idProduct, ent->d_name);
                strcpy(out_ttyname, ent->d_name);
                break;
            } 
        }
        closedir(pDir); 
    }
  
    return (out_ttyname[0] ? 0: -1);
}

int ql_get_ttyAT(char *out_ttyname) {
    int ret = ql_get_ttyname(USB_AT_INF, out_ttyname);
    if(ret != 0) {
        LOGE("cannot find ttyname for at");
        return -ENOENT;
    }
    return ret;
}

static int ql_pppd(int argc, char *argv[]) {
    int i = 0;
    char USER[PROPERTY_VALUE_MAX] = {'\0'};
    char PASSWORD[PROPERTY_VALUE_MAX] = {'\0'};
    char ttyname[20];
    pid_t PPPD_PID;

__ql_pppd_restart:
    ttyname[0] = 0;
    if (argc && (!strncmp(argv[0], "/dev", 4)))
        strncpy(ttyname, argv[0], sizeof(ttyname)-1);
    if(!ttyname[0] && ql_get_ttyname(USB_PPP_INF, ttyname) != 0) {
        LOGE("cannot find ttyname for ppp");
        property_set("net.gprs.ppp-exit", "2");
        ql_ip_down();
        return ENOENT;
    }

    property_set("net.gprs.ppp-pid", "-1");
    PPPD_PID = fork();
    if (0 == PPPD_PID) { //this is the child_process
        i = 0;
        while(pppd_args[i]) {
                if (0 == strcmp(pppd_args[i], "$USER")) {
                    property_get("net.gprs.user", (char *)&USER, "dummy");
                    pppd_args[i] = USER;
                } else  if (0 == strcmp(pppd_args[i], "$PASSWORD")) {
                    property_get("net.gprs.password", (char *)&PASSWORD, "dummy");
                    pppd_args[i] = PASSWORD;
                } else  if (0 == strcmp(pppd_args[i], "$DEVNAME")) {
                    pppd_args[i] = ttyname;
                }
                i++;
        }
    
        if (execv("/system/bin/pppd", (char**) pppd_args)) {
            LOGE("cannot execve('%s'): %s\n", "pppd", strerror(errno));
            return errno;
        }
        return 0;
    } else if (PPPD_PID < 0) {
        LOGE("failed to start ('%s'): %s\n", "pppd", strerror(errno));
        return errno;
    } else {
        int status, retval = 0;
        char str[PROPERTY_VALUE_MAX];
        sprintf(str, "%d", PPPD_PID);
        property_set("net.gprs.ppp-pid", str);
        waitpid(PPPD_PID, &status, 0);
        property_set("net.gprs.ppp-pid", "-1");
        if (WIFSIGNALED(status)) {
            LOGD("*** %s: Killed by signal %d\n", "pppd", WTERMSIG(status));
            retval = WTERMSIG(status);
        } else if (WIFEXITED(status) && WEXITSTATUS(status) > 0) {
            LOGD("*** %s: Exit code %d\n", "pppd", WEXITSTATUS(status));
            retval = WEXITSTATUS(status);
        }        
        sprintf(str, "%d", retval);
        property_set("net.gprs.ppp-exit", str);
        ql_ip_down();
        //pppd exit maybe becasue pppd error, or by ril-daemon.
        //if the reason is ril-daemon.
#define EXIT_USER_REQUEST	5
		if (EXIT_USER_REQUEST == retval)
			return 0;
        //if the reason is pppd error, to restart pppd again.
        while (1) {
            sleep(2);
            property_get("net.gprs.gprsState", str, "0");
            if (str[0] == '0')
                LOGD("gprsState is out of service!");
            else
                break;
        }
        goto __ql_pppd_restart;
    }

    return 0;     
}

#define ERROR_MAYBE_PPP_EXISTS -444
static int send_at(const char *at, const char *expect, int timeout) {
    struct pollfd poll_fd = {0, POLLIN, 0};
    int ret;
    char buf[128];
    const char *ABORT[] = {"BUSY", "NO CARRIER", "ERROR", "+CME ERROR:", "NO DIALTONE", "NO ANSWER"};
    size_t i;

    LOGD("chat --> %s", at);
    do {
        ret = write(1, at, strlen(at));
    } while (ret < 0 && errno == EINTR);
    
    if (ret <= 0) {
        LOGD("chat write error on stdout: %s(%d) ", strerror(errno), errno);
        return -1;
    }

    if(!expect)
        return 0;

    while(timeout > 0) {
        poll_fd.fd = 0;
        poll_fd.events = POLLIN;
        poll_fd.revents = 0;
        if(poll(&poll_fd, 1, 500) <= 0) {
            if (errno == ETIMEDOUT) {
                timeout -= 500;
                continue;
            } else if(errno != EINTR) {
                LOGE("chat poll error on stdin: %s(%d) ", strerror(errno), errno);
                if (!errno)
                    return ERROR_MAYBE_PPP_EXISTS;
                return -1;
            }
        }
        
        if(poll_fd.revents && (poll_fd.revents & POLLIN)) {
            memset(buf, 0, sizeof(buf));
            usleep(100*1000);
            if(read(0, buf, sizeof(buf)-1) < 0) {
                LOGD("chat read error on stdin: %s(%d) ", strerror(errno), errno);
                return -1;
            }
            LOGD("chat <-- %s", buf);
            for(i = 0; i < ARRAY_SIZE(ABORT); i++)
                if(strstr(buf, ABORT[i]))
                    return -1;  
            if(strstr(buf, expect)) {
                return 0;
            }
        }    
    }

    return -1;
}

static int ql_chat(int argc, char *argv[]) {
    int flags;
    int ret = -1, i;

    if ((flags = fcntl(0, F_GETFL, 0)) == -1) {
        LOGD("chat Can't get file mode flags on stdin: %s(%d) ", strerror(errno), errno);
        return -1;
    }

    if (fcntl(0, F_SETFL, flags | O_NONBLOCK) == -1) {
        LOGD("chat Can't set file mode flags on stdin: %s(%d) ", strerror(errno), errno);
        return -1;
    }

    if (!strcmp(argv[0], "disconnect")) {
        // 3 choice to terminate Data Call, 1: send SIGTERM to pppd, 2: below code, 3: send "AT+CGACT=0,1" via at channel
        sleep(1);
        ret = send_at("+++", 0, 0);
        //sleep(1);
    } else if(!strcmp(argv[0], "connect")) {
        for (i = 0; i < 2; i++) {
            ret = send_at("AT\r", "OK", 3*1000);
            if (ret != 0) {
                if (ret == ERROR_MAYBE_PPP_EXISTS) {
                    sleep(1);
                    send_at("+++", 0, 0);
                    sleep(3);
                }
                ret = send_at("AT\r", "OK", 3*1000);
            }
                          
            if (ret == 0) {
                ret = send_at("ATD*99#\r", "CONNECT", 10*1000);
                if (ret == 0)
                    break;
                else if (i == 0) { //maybe in data mode
                    send_at("AT+CGACT=0,1\r", "OK", 3*1000);
                    sleep(3);
                }
            } else {
                sleep(1);
            }
        }
    }
       
    return ret;
}

static void ql_sigaction(int signo) {
    switch (signo) {
    case SIGINT:
        LOGE("SIGINT");
    break;
    case SIGTERM:
        LOGE("SIGTERM");
    break;
    case SIGHUP:
        LOGE("SIGHUP");
    break;
    default:
        LOGE("chat signo = %d", signo);
    break;
    }
}

#define MAX_LINE_SIZE 512
#define MAX_RESULT 20
static char *qitems[MAX_RESULT];
static char *qout;
static char * newline;
static char * qdelspace(char *oldline) {
    char *pnewline = newline;

    while ((*oldline == ' ') || (*oldline == '\t'))
        oldline++;

    while (*oldline) {
        if ((*oldline == ' ') || (*oldline == '\t')) {
            if ((oldline[1] != ' ') && (oldline[1] != '\t'))
                *pnewline++ = ' ';
            oldline++;
        } else {
            *pnewline++ = *oldline++;            
        }
    }

    *pnewline++ = '\0'; 
    return newline;
}

static int qgrep(const char *cmd, const char *key[], int keynum) {
    FILE *fp;
    int findnum = 0;
    char line[MAX_LINE_SIZE]= "\n";
    int i = 0;

    strcat(line, cmd);
    if (keynum)
        strcat(line, " | grep ");
    for (i = 0; i < keynum; i++) {
        strcat(line, key[i]);
        strcat(line, " ");
    }
    printf("%s\n", line);
   
    fp = popen(cmd, "r");
    if (!fp) {
        printf("exec '%s' error : %d (%s)\n", cmd, errno, strerror(errno));
        return 0;
    }

    qitems[0] = qout;    
    while((fgets(line, sizeof(line), fp)) != NULL) {
        int find  = 0;
        if (!keynum) {
            find = 1;          
        } else {
            for (i = 0; i < keynum; i++) {
                if (strstr(line, key[i])) {
                    find = 1;
                    break;
                }
            }
        }

        if (find) {
            char *pend = line + strlen(line) - 1;
            while ((*pend == '\r') || (*pend == '\n') || (*pend == ' ') || (*pend == '\t'))
                *pend-- = '\0';
            strcpy(qitems[findnum], line);
            qitems[findnum+1] = qitems[findnum] + strlen(qitems[findnum]) + 1;
            printf("%s\n", qitems[findnum]);
            findnum++;
        }
    }

    pclose(fp);
    return findnum;
}

static int quectel_debug(void) {
    int num, i;
    const char *key[5];
    const char *cmd;
    int fail = 0;
    int pid = 0;

    newline = (char *)malloc(MAX_LINE_SIZE);
    qout = (char *)malloc(MAX_RESULT*MAX_LINE_SIZE);
    
//check service ril-daemon
    cmd = "getprop init.svc.ril-daemon";
    for (i = 0; i < 2; i++) {
        num = qgrep(cmd, 0, 0);
        if (!num || !strcmp(qitems[0], "")) {
            printf("Couldn't find service ril-daemon (defined in init.rc?)\n");
            fail = 1;
            break;
        } else if (strcmp(qitems[0], "running")) {
            printf("service ril-daemon is %s (shoule be running)\n", qitems[0]);
            fail = 1;
            break;
        }
        usleep(50*1000);
    }
    
    if (fail) {
        cmd = "cat /init.rc";
        key[0] = "ril-daemon";
        fail = 1;
        num = qgrep(cmd, key, 1);
        if (num) {
            const char *service_ril = "service ril-daemon /system/bin/rild -l /system/lib/libreference-ril.so";
            for (i = 0; i < num; i++) {
                char * pnewline = qdelspace(qitems[i]);
                if ((pnewline[0] == '#') || strstr(pnewline, "stop") ||  strstr(pnewline, "start"))
                    continue;
                if (strncmp(pnewline, service_ril, strlen(service_ril))) {
                    printf("%s\t is wrong\n", qitems[i]);
                } else {
                    fail = 0;
                }
            }
        } 

        if (fail)
            printf("Couldn't find service ril-daemon in init.rc\n");
    }else {
    //check rild's user shoud be root, not radio
        cmd = "ps";
        key[0] = "/system/bin/rild";
        num = qgrep(cmd, key, 1);
        for (i = 0; i < num; i++) {
            if (strstr(qitems[i], "radio")) {
                printf("do you comment out switchUser() in hardware/ril/rild/rild.c\n");
            }
        }
    }

// check ril version? is proviedy by quectel?
    cmd = "getprop gsm.version.ril-impl";
    num = qgrep(cmd, 0, 0);
    if (!num || !strcmp(qitems[0], "")) {
        printf("Couldn't getprop gsm.version.ril-impl, service ril-daemon is running?\n");
    } else {
        if (strstr(qitems[0], "Quectel") || strstr(qitems[0], "quectel"))
            ;
        else
            printf("RIL is not provided by Quectel\n");
    }

//check power on moderm?  configure usb drive? baseband version?
    cmd = "getprop gsm.version.baseband";
    num = qgrep(cmd, 0, 0);
    if (!num || !strcmp(qitems[0], "")) {
        int usb_pid = 0;
        char out_ttyname[20];
        printf("Couldn't getprop gsm.version.baseband\n");
        if (ql_get_ttyname(USB_AT_INF, out_ttyname) != 0) {
            usb_pid = ql_get_ttyname(0xff, out_ttyname);
            cmd = "ls -l /sys/bus/usb/drivers"; 
            key[0] = '\0';
            if ((usb_pid == 0x9003)  || (usb_pid == 0x9090) || (usb_pid == 0x9215))
                key[0] = "option";
            else if ((usb_pid == 0x0331) || (usb_pid == 0x0020))
                key[0] = "cdc_acm";
            if (key[0]) {
                num = qgrep(cmd, key, 1);
                if (!num)
                    printf("Couldn't find usb driver (configured in linux kernel?)\n");
            } else {
                printf("Couldn't find Module (have power on?)\n");
            }
        }
    }

// check /etc/ppp/ip-up /etc/ppp/ip-down
    cmd = "ls -l /etc/ppp/ip-up"; 
    key[0] = "ip-up";
    num = qgrep(cmd, key, 1);
    if (!num || !strstr(qitems[0], "init.quectel-pppd")) {
        printf("Couldn't find /etc/ppp/ip-up (be link file, linked to /etc/init.quectel-pppd, generated by Quectel ril)\n");
    }
    cmd = "ls -l /etc/ppp/ip-down"; 
    key[0] = "ip-down";
    num = qgrep(cmd, key, 1);
    if (!num || !strstr(qitems[0], "init.quectel-pppd")) {
        printf("Couldn't find /etc/ppp/ip-down (be link file, linked to /etc/init.quectel-pppd, generated by Quectel ril)\n");
    }
 
//check service quectel-pppd
    cmd = "getprop init.svc.quectel-pppd";
    num = qgrep(cmd, 0, 0);
    if (!num || !strcmp(qitems[0], "")) {
        printf("Couldn't find service quectel-pppd\n");
        cmd = "cat /init.rc";
        key[0] = "quectel-pppd";
        fail = 1;
        num = qgrep(cmd, key, 1);
        if (num) {
            const char *service_pppd = "service quectel-pppd /etc/ppp/init.quectel-pppd";
            for (i = 0; i < num; i++) {
                char * pnewline = qdelspace(qitems[i]);
                if ((pnewline[0] == '#') || strstr(pnewline, "stop") ||  strstr(pnewline, "start"))
                    continue;
                if (strncmp(pnewline, service_pppd, strlen(service_pppd))) {
                    printf("%s\t  is wrong\n", qitems[i]);
                } else {
                    fail = 0;
                }
            }
        }    
        if (fail)
            printf("Couldn't find service quectel-pppd in init.rc\n");
    }

//check APN setting
    cmd = "getprop net.ppp0.user";
    num = qgrep(cmd, 0, 0);
    if (!num || !strcmp(qitems[0], "")) {
        printf("Couldn't find APN (Settings -> WIRELESS & NETWORKS -> More... -> Mobile networks -> Access Point Names)\n");
    }

//check data sevice state
    cmd = "getprop net.gprs.gprsState";
    num = qgrep(cmd, 0, 0);
    if (!num  || !strcmp(qitems[0], ""))
        ;
    else if (!strcmp(qitems[0], "0")) {
        printf("Couldn't register Data Service!\n");
    }    
//check data sevice state
    cmd = "ls -l /system/bin/pppd";
    num = qgrep(cmd, 0, 0);
    if (!num)
        printf("Couldn't /system/bin/pppd\n");
    
//show dmesg
    cmd = "dmesg";
    key[0] = "quectel-pppd";
    key[1] = "ril-daemon";
    num = qgrep(cmd, key, 2);
    for (i = 0; i < num; i++) {
        if (strstr(qitems[i], "no such service")) {
            printf("do you define service quectel-pppd in init.rc\n");
            break;
        } else if (strstr(qitems[i], "sys_prop: Unable to")) {
            printf("do you comment out switchUser() in hardware/ril/rild/rild.c\n");
            break;
        }
    }

    free(newline);
    free(qout);
    printf("\n");
    return 0;
}

static char quectel_argv[512];
int main(int argc, char *argv[]) {
    int error = -EEXIST;
    int i = 0;

    signal(SIGINT, ql_sigaction);
    signal(SIGTERM, ql_sigaction);
    signal(SIGHUP, ql_sigaction);

    quectel_argv[0] = 0;
    for (i = 0; i < argc; i++) {
        strcat(quectel_argv, argv[i]);
        strcat(quectel_argv, " ");
   }
    LOGD("%s", quectel_argv);

    if (strstr(argv[0], "init.quectel-pppd")) {
        if ((argc == 3) && (!strcmp(argv[1], "chat")))
            error  = ql_chat(argc - 2, argv + 2);
        else if ((argc == 2) && (!strcmp(argv[1], "debug")))
            error = quectel_debug();
        else
            error = ql_pppd(argc - 1, argv + 1);
    } else if (strstr(argv[0], "ip-up")) {
        error = ql_ip_up();
    } else if (strstr(argv[0], "ip-down")) {
        error = ql_ip_down();
    } else {
        LOGE("unknow commad (%s)", argv[0]);
        error = -EEXIST;
    }

    return error;
}
